package SevenZip.Compression.LZMA;

import SevenZip.ICodeProgress;
import SevenZip.InvalidParamException;
import SevenZip.Compression.LZ.OutWindow;
import SevenZip.Compression.LZMA.Base;
import SevenZip.Compression.RangeCoder.BitDecoder;
import SevenZip.Compression.RangeCoder.BitTreeDecoder;

public class Decoder /* implements ICoder, ISetDecoderProperties */ {
    class LenDecoder {
        BitDecoder m_Choice = new BitDecoder();
        BitDecoder m_Choice2 = new BitDecoder();
        BitTreeDecoder[] m_LowCoder = new BitTreeDecoder[Base.kNumPosStatesMax];
        BitTreeDecoder[] m_MidCoder = new BitTreeDecoder[Base.kNumPosStatesMax];
        BitTreeDecoder m_HighCoder = new BitTreeDecoder(Base.kNumHighLenBits);
        int m_NumPosStates = 0;
        
        public void Create(int numPosStates) {
            for (int posState = m_NumPosStates; posState < numPosStates; posState++) {
                m_LowCoder[posState] = new BitTreeDecoder(Base.kNumLowLenBits);
                m_MidCoder[posState] = new BitTreeDecoder(Base.kNumMidLenBits);
            }
            m_NumPosStates = numPosStates;
        }
        
        public void Init() {
            m_Choice.Init();
            for (int posState = 0; posState < m_NumPosStates; posState++) {
                m_LowCoder[posState].Init();
                m_MidCoder[posState].Init();
            }
            m_Choice2.Init();
            m_HighCoder.Init();
        }
        
        public final int Decode(SevenZip.Compression.RangeCoder.Decoder rangeDecoder, int posState) throws java.io.IOException {
            if (m_Choice.Decode(rangeDecoder) == 0)
                return m_LowCoder[posState].Decode(rangeDecoder);
            else {
                int symbol = Base.kNumLowLenSymbols;
                if (m_Choice2.Decode(rangeDecoder) == 0)
                    symbol += m_MidCoder[posState].Decode(rangeDecoder);
                else {
                    symbol += Base.kNumMidLenSymbols;
                    symbol += m_HighCoder.Decode(rangeDecoder);
                }
                return symbol;
            }
        }
    }
    
    class LiteralDecoder {
        class Decoder2 {
            BitDecoder[] m_Decoders;
            public void Create() {
                m_Decoders = new BitDecoder[0x300];
                for (int i = 0; i < 0x300; i++) {
                    m_Decoders[i] = new BitDecoder();
                }
            }
            public void Init() {
                for (int i = 0; i < 0x300; i++) {
                    m_Decoders[i].Init();
                }
            }
            
            public int DecodeNormal(SevenZip.Compression.RangeCoder.Decoder rangeDecoder) throws java.io.IOException {
                int symbol = 1;
                do
                    symbol = (symbol << 1) | m_Decoders[symbol].Decode(rangeDecoder);
                while (symbol < 0x100);
                return symbol;
            }
            
            public final int DecodeWithMatchByte(SevenZip.Compression.RangeCoder.Decoder rangeDecoder, int matchByte) throws java.io.IOException {
                int symbol = 1;
                do
                {
                    int matchBit = (matchByte >> 7) & 1;
                    matchByte <<= 1;
                    int bit = m_Decoders[((1 + matchBit) << 8) + symbol].Decode(rangeDecoder);
                    symbol = (symbol << 1) | bit;
                    if (matchBit != bit) {
                        while (symbol < 0x100)
                            symbol = (symbol << 1) | m_Decoders[symbol].Decode(rangeDecoder);
                        break;
                    }
                }
                while (symbol < 0x100);
                return symbol;
            }
        }
        
        Decoder2[] m_Coders;
        int m_NumPrevBits;
        int m_NumPosBits;
        int m_PosMask;
        
        public final void Create(int numPosBits, int numPrevBits) {
            if (m_Coders != null && m_NumPrevBits == numPrevBits &&
                    m_NumPosBits == numPosBits)
                return;
            m_NumPosBits = numPosBits;
            m_PosMask = (1 << numPosBits) - 1;
            m_NumPrevBits = numPrevBits;
            int numStates = 1 << (m_NumPrevBits + m_NumPosBits);
            m_Coders = new Decoder2[numStates];
            for (int i = 0; i < numStates; i++) {
                m_Coders[i] = new Decoder2();
                m_Coders[i].Create();
            }
        }
        
        public final void Init() {
            int numStates = 1 << (m_NumPrevBits + m_NumPosBits);
            for (int i = 0; i < numStates; i++)
                m_Coders[i].Init();
        }
        
        final int GetState(int pos, int prevByte) {
            return ((pos & m_PosMask) << m_NumPrevBits) + (prevByte >>> (8 - m_NumPrevBits)); }
        
        public final int DecodeNormal(SevenZip.Compression.RangeCoder.Decoder rangeDecoder, int pos, int prevByte) throws java.io.IOException {
            return m_Coders[GetState(pos, prevByte)].DecodeNormal(rangeDecoder); }
        
        public final int DecodeWithMatchByte(SevenZip.Compression.RangeCoder.Decoder rangeDecoder, int pos, int prevByte, int matchByte) throws java.io.IOException {
            return m_Coders[GetState(pos, prevByte)].DecodeWithMatchByte(rangeDecoder, matchByte); }
    }
    
    
    OutWindow m_OutWindow = new OutWindow();
    SevenZip.Compression.RangeCoder.Decoder m_RangeDecoder = new SevenZip.Compression.RangeCoder.Decoder();
    
    BitDecoder[] m_IsMatchDecoders = new BitDecoder[Base.kNumStates << Base.kNumPosStatesBitsMax];
    BitDecoder[] m_IsRepDecoders = new BitDecoder[Base.kNumStates];
    BitDecoder[] m_IsRepG0Decoders = new BitDecoder[Base.kNumStates];
    BitDecoder[] m_IsRepG1Decoders = new BitDecoder[Base.kNumStates];
    BitDecoder[] m_IsRepG2Decoders = new BitDecoder[Base.kNumStates];
    BitDecoder[] m_IsRep0LongDecoders = new BitDecoder[Base.kNumStates << Base.kNumPosStatesBitsMax];
    
    BitTreeDecoder[] m_PosSlotDecoder = new BitTreeDecoder[Base.kNumLenToPosStates];
    BitDecoder[] m_PosDecoders = new BitDecoder[Base.kNumFullDistances - Base.kEndPosModelIndex];
    
    BitTreeDecoder m_PosAlignDecoder = new BitTreeDecoder(Base.kNumAlignBits);
    
    LenDecoder m_LenDecoder = new LenDecoder();
    LenDecoder m_RepLenDecoder = new LenDecoder();
    
    LiteralDecoder m_LiteralDecoder = new LiteralDecoder();
    
    int m_DictionarySize;
    int m_DictionarySizeCheck;
    
    int m_PosStateMask;
    
    public Decoder() {
        m_DictionarySize = 0xFFFFFFFF;
        int i;
        for (i = 0; i < Base.kNumLenToPosStates; i++)
            m_PosSlotDecoder[i] = new BitTreeDecoder(Base.kNumPosSlotBits);
        
        for (i = 0; i < Base.kNumStates; i++) {
            m_IsRepDecoders[i] = new BitDecoder();
            m_IsRepG0Decoders[i] = new BitDecoder();
            m_IsRepG1Decoders[i] = new BitDecoder();
            m_IsRepG2Decoders[i] = new BitDecoder();
        }
        for (i = 0; i < Base.kNumFullDistances - Base.kEndPosModelIndex; i++) {
            m_PosDecoders[i] = new BitDecoder();
        }
    }
    void SetDictionarySize(int dictionarySize) {
        if (m_DictionarySize != dictionarySize) {
            m_DictionarySize = dictionarySize;
            m_DictionarySizeCheck = Math.max(m_DictionarySize, 1);
            int blockSize = Math.max(m_DictionarySizeCheck, (1 << 12));
            m_OutWindow.Create(blockSize);
        }
    }
    
    void SetLiteralProperties(int lp, int lc) throws InvalidParamException {
        if (lp > 8)
            throw new InvalidParamException();
        if (lc > 8)
            throw new InvalidParamException();
        m_LiteralDecoder.Create(lp, lc);
    }
    
    void SetPosBitsProperties(int pb)  throws InvalidParamException {
        if (pb > Base.kNumPosStatesBitsMax)
            throw new InvalidParamException();
        int numPosStates = 1 << pb;
        m_LenDecoder.Create(numPosStates);
        m_RepLenDecoder.Create(numPosStates);
        m_PosStateMask = numPosStates - 1;
    }
    
    public void SetDecoderProperties(byte[] properties) throws InvalidParamException {
        if (properties.length < 5)
            throw new InvalidParamException();
        int prop0 = properties[0] & 0xFF;
        int lc = prop0 % 9;
        int remainder = prop0 / 9;
        int lp = remainder % 5;
        int pb = remainder / 5;
        if (pb > Base.kNumPosStatesBitsMax)
            throw new InvalidParamException();
        int dictionarySize = 0;
        for (int i = 0; i < 4; i++)
            dictionarySize += (properties[1 + i] & 0xFF) << (i * 8);
        SetDictionarySize(dictionarySize);
        SetLiteralProperties(lp, lc);
        SetPosBitsProperties(pb);
    }
    void Init(java.io.InputStream inStream, java.io.OutputStream outStream) throws java.io.IOException {
        m_RangeDecoder.Init(inStream);
        m_OutWindow.Init(outStream);
        
        int i;
        for (i = 0; i < Base.kNumStates; i++) {
            for (int j = 0; j <= m_PosStateMask; j++) {
                int index = (i << Base.kNumPosStatesBitsMax) + j;
                m_IsMatchDecoders[index] = new BitDecoder();
                m_IsMatchDecoders[index].Init();
                m_IsRep0LongDecoders[index] = new BitDecoder();
                m_IsRep0LongDecoders[index].Init();
            }
            m_IsRepDecoders[i].Init();
            m_IsRepG0Decoders[i].Init();
            m_IsRepG1Decoders[i].Init();
            m_IsRepG2Decoders[i].Init();
        }
        
        m_LiteralDecoder.Init();
        for (i = 0; i < Base.kNumLenToPosStates; i++)
            m_PosSlotDecoder[i].Init();
        // m_PosSpecDecoder.Init();
        for (i = 0; i < Base.kNumFullDistances - Base.kEndPosModelIndex; i++) {
            m_PosDecoders[i].Init();
        }
        
        m_LenDecoder.Init();
        m_RepLenDecoder.Init();
        m_PosAlignDecoder.Init();
    }
    
    public void Code(java.io.InputStream inStream, java.io.OutputStream outStream,long outSize, ICodeProgress progress) throws java.io.IOException {
        Init(inStream, outStream);
        
        int stateIndex = 0;
        
        boolean previousIsMatch = false;
        int rep0 = 0, rep1 = 0, rep2 = 0, rep3 = 0;
        
        long nowPos64 = 0;
        long outSize64 = outSize;
        if (nowPos64 < outSize64) {
            if (m_IsMatchDecoders[stateIndex << Base.kNumPosStatesBitsMax].Decode(m_RangeDecoder) != 0)
                throw new java.io.IOException("DataErrorException-1");
            stateIndex = Base.UpdateChar(stateIndex);
            int b = m_LiteralDecoder.DecodeNormal(m_RangeDecoder, 0, 0);
            m_OutWindow.PutByte(b);
            nowPos64++;
        }
        while (nowPos64 < outSize64) {
            // UInt64 next = Math.Min(nowPos64 + (1 << 18), outSize64);
            // while(nowPos64 < next)
            {   
                int posState = (int)nowPos64 & m_PosStateMask;             
                if (m_IsMatchDecoders[(stateIndex << Base.kNumPosStatesBitsMax) + posState].Decode(m_RangeDecoder) == 0) {
                    stateIndex = Base.UpdateChar(stateIndex);
                    int b;
                    int prevByte = m_OutWindow.GetByte(0);
                    if (previousIsMatch) {
                        b = m_LiteralDecoder.DecodeWithMatchByte(m_RangeDecoder,
                                (int)nowPos64, prevByte,
                                m_OutWindow.GetByte(rep0));
                        previousIsMatch = false;
                    } else
                        b = m_LiteralDecoder.DecodeNormal(m_RangeDecoder,
                                (int)nowPos64, prevByte);
                    m_OutWindow.PutByte(b);
                    nowPos64++;
                } else {
                    previousIsMatch = true;
                    int len;                    
                    if (m_IsRepDecoders[stateIndex].Decode(m_RangeDecoder) == 1) {   
                        if (m_IsRepG0Decoders[stateIndex].Decode(m_RangeDecoder) == 0) {   
                            if (m_IsRep0LongDecoders[(stateIndex << Base.kNumPosStatesBitsMax) + posState].Decode(m_RangeDecoder) == 0) { 
                                stateIndex = Base.UpdateShortRep(stateIndex);
                                m_OutWindow.PutByte(m_OutWindow.GetByte(rep0));
                                nowPos64++;
                                continue;
                            }
                        } else {    
                            int distance;
                            if (m_IsRepG1Decoders[stateIndex].Decode(m_RangeDecoder) == 0) {
                                distance = rep1;
                            } else {
                                if (m_IsRepG2Decoders[stateIndex].Decode(m_RangeDecoder) == 0)
                                    distance = rep2;
                                else {
                                    distance = rep3;
                                    rep3 = rep2;
                                }
                                rep2 = rep1;
                            }
                            rep1 = rep0;
                            rep0 = distance;
                        }
                        len = m_RepLenDecoder.Decode(m_RangeDecoder, posState) + Base.kMatchMinLen;
                        stateIndex = Base.UpdateRep(stateIndex);
                    } else {    
                        rep3 = rep2;
                        rep2 = rep1;
                        rep1 = rep0;
                        len = Base.kMatchMinLen + m_LenDecoder.Decode(m_RangeDecoder, posState);
                        stateIndex = Base.UpdateMatch(stateIndex);
                        int posSlot = m_PosSlotDecoder[Base.GetLenToPosState(len)].Decode(m_RangeDecoder);
                        if (posSlot >= Base.kStartPosModelIndex) {
                            int numDirectBits = (int)((posSlot >>> 1) - 1);
                            rep0 = ((2 | (posSlot & 1)) << numDirectBits);
                            if (posSlot < Base.kEndPosModelIndex)
                                rep0 += BitTreeDecoder.ReverseDecode(m_PosDecoders,
                                        rep0 - posSlot - 1, m_RangeDecoder, numDirectBits);
                            else {
                                rep0 += (m_RangeDecoder.DecodeDirectBits(
                                        numDirectBits - Base.kNumAlignBits) << Base.kNumAlignBits);
                                rep0 += m_PosAlignDecoder.ReverseDecode(m_RangeDecoder);
                            }
                        } else
                            rep0 = posSlot;
                    }
                    if (rep0 >= nowPos64 || rep0 >= m_DictionarySizeCheck) {    
                        if (rep0 == 0xFFFFFFFF)
                            break;
                        throw new java.io.IOException("DataErrorException-2 "+rep0+" "+nowPos64+" "+m_DictionarySizeCheck);
                    }
                    m_OutWindow.CopyBlock(rep0, len);
                    nowPos64 += len;
                }
            }
        }
        m_OutWindow.Flush();
        m_OutWindow.ReleaseStream();
        m_RangeDecoder.ReleaseStream();
    }
}
