package SevenZip.Compression.RangeCoder;

public	class BitTreeEncoder {
    BitEncoder[] Models;
    int NumBitLevels;
    
    public BitTreeEncoder(int numBitLevels) {
        NumBitLevels = numBitLevels;
        Models = new BitEncoder[1 << numBitLevels];
        for (int i = 1; i < (1 << NumBitLevels); i++) {
            Models[i] = new BitEncoder();
        }
    }
    
    public void Init() {
        for (int i = 1; i < (1 << NumBitLevels); i++) {
            Models[i].Init();
        }
    }
    
    public void Encode(Encoder rangeEncoder, int symbol) throws java.io.IOException {
        int m = 1;
        for (int bitIndex = NumBitLevels; bitIndex > 0; ) {
            bitIndex--;
            int bit = (symbol >> bitIndex) & 1;
            Models[m].Encode(rangeEncoder, bit);
            m = (m << 1) | bit;
        }
    }
    
    public void ReverseEncode(Encoder rangeEncoder, int symbol) throws java.io.IOException {
        int m = 1;
        for (int i = 0; i < NumBitLevels; i++) {
            int bit = symbol & 1;
            Models[m].Encode(rangeEncoder, bit);
            m = (m << 1) | bit;
            symbol >>= 1;
        }
    }
    
    public int GetPrice(int symbol) {
        int price = 0;
        int m = 1;
        for (int bitIndex = NumBitLevels; bitIndex > 0; ) {
            bitIndex--;
            int bit = (symbol >> bitIndex) & 1;
            price += Models[m].GetPrice(bit);
            m = (m << 1) + bit;
        }
        return price;
    }
    
    public int ReverseGetPrice(int symbol) {
        int price = 0;
        int m = 1;
        for (int i = NumBitLevels; i > 0; i--) {
            int bit = symbol & 1;
            symbol >>= 1;
            price += Models[m].GetPrice(bit);
            m = (m << 1) | bit;
        }
        return price;
    }
    
    public static int ReverseGetPrice(BitEncoder[] Models, int startIndex, int NumBitLevels, int symbol) {
        int price = 0;
        int m = 1;
        for (int i = NumBitLevels; i > 0; i--) {
            int bit = symbol & 1;
            symbol >>= 1;
            price += Models[startIndex + m].GetPrice(bit);
            m = (m << 1) | bit;
        }
        return price;
    }
    
    public static void ReverseEncode(BitEncoder[] Models, int startIndex,
            Encoder rangeEncoder, int NumBitLevels, int symbol) throws java.io.IOException {
        int m = 1;
        for (int i = 0; i < NumBitLevels; i++) {
            int bit = symbol & 1;
            Models[startIndex + m].Encode(rangeEncoder, bit);
            m = (m << 1) | bit;
            symbol >>= 1;
        }
    }
}
